/*****************************************************************************
 *                                 jugCLer                                   *
 *              realtime raytracing experiment: setup and glue               *
 *                                                                           *
 *  Copyright (C) 2013  Holger Bettag               (hobold@vectorizer.org)  *
 *  Parts of this code are based on code from David Bucciarelli's MandelGPU  *
 *  program. It is being used here with David's permission.                  *
 *                                                                           *
 *  This program is free software; you can redistribute it and/or modify     *
 *  it under the terms of the GNU General Public License as published by     *
 *  the Free Software Foundation; either version 2 of the License, or        *
 *  (at your option) any later version.                                      *
 *                                                                           *
 *  This program is distributed in the hope that it will be useful,          *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *  GNU General Public License for more details.                             *
 *                                                                           *
 *  You should have received a copy of the GNU General Public License along  *
 *  with this program; if not, write to the Free Software Foundation, Inc.,  *
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.              *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef WIN32
  #include <GLUT/glut.h>
  #include <CL/opencl.h>
#else
  #ifdef __APPLE__
    #include <GLUT/glut.h>
    #include <OpenCL/opencl.h>
  #else   // linux
    #include <GL/glut.h>
    #include <CL/cl.h>
  #endif
#endif

#include "utils.h"      // bitmap, timing
#include "scene.h"      // data structures describing scenes
#include "animation.h"  // functions to animate the juggler


/**********************************
 *  simple OpenCL error handling  *
 **********************************/
cl_int status;  // global error variable, unsafe but convenient for a start

const char* openCLerrorNumberToString(int error) {
  switch (error) {
  case CL_SUCCESS: return "Success";
  case CL_DEVICE_NOT_FOUND: return "Device not found";
  case CL_DEVICE_NOT_AVAILABLE: return "Device not available";
  case CL_COMPILER_NOT_AVAILABLE: return "Compiler not available";
  case CL_MEM_OBJECT_ALLOCATION_FAILURE:
    return "Memory object allocation failure";
  case CL_OUT_OF_RESOURCES: return "Out of resources";
  case CL_OUT_OF_HOST_MEMORY: return "Out of host memory";
  case CL_PROFILING_INFO_NOT_AVAILABLE:
    return "Profiling information not available";
  case CL_MEM_COPY_OVERLAP: return "Memory copy overlap";
  case CL_IMAGE_FORMAT_MISMATCH: return "Image format mismatch";
  case CL_IMAGE_FORMAT_NOT_SUPPORTED: return "Image format not supported";
  case CL_BUILD_PROGRAM_FAILURE: return "Program build failure";
  case CL_MAP_FAILURE: return "Map failure";
  case CL_MISALIGNED_SUB_BUFFER_OFFSET: return "Misaligned sub-buffer offset";
  case CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST:
    return "Execution status error for events in wait list";

  case CL_INVALID_VALUE: return "Invalid value";
  case CL_INVALID_DEVICE_TYPE: return "Invalid device type";
  case CL_INVALID_PLATFORM: return "Invalid platform";
  case CL_INVALID_DEVICE: return "Invalid device";
  case CL_INVALID_CONTEXT: return "Invalid context";
  case CL_INVALID_QUEUE_PROPERTIES: return "Invalid queue properties";
  case CL_INVALID_COMMAND_QUEUE: return "Invalid command queue";
  case CL_INVALID_HOST_PTR: return "Invalid host pointer";
  case CL_INVALID_MEM_OBJECT: return "Invalid memory object";
  case CL_INVALID_IMAGE_FORMAT_DESCRIPTOR:
    return "Invalid image format descriptor";
  case CL_INVALID_IMAGE_SIZE: return "Invalid image size";
  case CL_INVALID_SAMPLER: return "Invalid sampler";
  case CL_INVALID_BINARY: return "Invalid binary";
  case CL_INVALID_BUILD_OPTIONS: return "Invalid build options";
  case CL_INVALID_PROGRAM: return "Invalid program";
  case CL_INVALID_PROGRAM_EXECUTABLE: return "Invalid program executable";
  case CL_INVALID_KERNEL_NAME: return "Invalid kernel name";
  case CL_INVALID_KERNEL_DEFINITION: return "Invalid kernel definition";
  case CL_INVALID_KERNEL: return "Invalid kernel";
  case CL_INVALID_ARG_INDEX: return "Invalid argument index";
  case CL_INVALID_ARG_VALUE: return "Invalid argument value";
  case CL_INVALID_ARG_SIZE: return "Invalid argument size";
  case CL_INVALID_KERNEL_ARGS: return "Invalid kernel arguments";
  case CL_INVALID_WORK_DIMENSION: return "Invalid work dimension";
  case CL_INVALID_WORK_GROUP_SIZE: return "Invalid work group size";
  case CL_INVALID_WORK_ITEM_SIZE: return "Invalid work item size";
  case CL_INVALID_GLOBAL_OFFSET: return "Invalid global offset";
  case CL_INVALID_EVENT_WAIT_LIST: return "Invalid event wait list";
  case CL_INVALID_EVENT: return "Invalid event";
  case CL_INVALID_OPERATION: return "Invalid operation";
  case CL_INVALID_GL_OBJECT: return "Invalid OpenGL object";
  case CL_INVALID_BUFFER_SIZE: return "Invalid buffer size";
  case CL_INVALID_MIP_LEVEL: return "Invalid mip-map level";
  case CL_INVALID_GLOBAL_WORK_SIZE: return "Invalid global work size";
  case CL_INVALID_PROPERTY: return "Invalid property";
  default: return "OOPS!! UNKNOWN OPENCL ERROR";
  }
  return "OOPS!! UNHANDLED CASE IN SWITCH";
}

// better than nothing error check and report
void reportOpenCLerror(int line, const char* msg) {
  if (status != CL_SUCCESS) {
    const char* errorName = openCLerrorNumberToString(status);
    fprintf(stderr, "line %d: %s\n", line, msg);
    fprintf(stderr, "(CL error number %d: %s)\n", status, errorName);
    exit(EXIT_FAILURE);
  }
}

////////////////////////////////////////////////////////////////////////////////
// GLUT, which we use for displaying images, makes use of callback functions. //
// The following global variables are the simplest way of passing parameters  //
// between the various callback functions.                                    //
////////////////////////////////////////////////////////////////////////////////

/*
const int imgWidth = 1568;
const int imgHeight = 980;
*/

const int imgWidth = 1024;
const int imgHeight = 640;

/*
const int imgWidth = 1280;
const int imgHeight = 800;
*/
Bitmap* bitmap = 0;          // image for display

Scene* scene;                // contains camera, light, and spheres

float throughput;     // passes performance info from measurement to display

///////////////////////
// OpenCL state info //
///////////////////////
cl_context context;
cl_device_id *devices = NULL;
cl_mem d_pixelBuffer;
cl_mem d_scene;
cl_command_queue commandQueue;
cl_program program;
cl_kernel kernel;
const char *kernelFileName = "trace.cl";

unsigned int workGroupSize = 1;

/******************************************
 *  functions used during GLUT callbacks  *
 ******************************************/

// compute and draw image
void computeImage(Scene* scene, PixelRGBA8888* Image) {
  // ring buffer for computation times of last 32 frames
  static cl_ulong times[32] =
    {1000000000, 1000000000, 1000000000, 1000000000,
     1000000000, 1000000000, 1000000000, 1000000000,
     1000000000, 1000000000, 1000000000, 1000000000,
     1000000000, 1000000000, 1000000000, 1000000000,
     1000000000, 1000000000, 1000000000, 1000000000,
     1000000000, 1000000000, 1000000000, 1000000000,
     1000000000, 1000000000, 1000000000, 1000000000,
     1000000000, 1000000000, 1000000000, 1000000000};
  static int timeIdx = 0;  // index into ring buffer
  static cl_ulong total_time = 32000000000UL;

  cl_event theEvent, kernelEvent;     // for synchronization & timing

  // copy scene from host to device
  status = clEnqueueWriteBuffer(commandQueue,
				d_scene,
				CL_FALSE,
				0,
				sizeof(Scene),
				scene,
				0,
				NULL,
				&theEvent);
  reportOpenCLerror(__LINE__, "Failed to write to OpenCL input buffer");
  // wait for the copy operation to finish
  status = clWaitForEvents(1, &theEvent);
  reportOpenCLerror(__LINE__, "Failed to wait for write to OpenCL input");
  clReleaseEvent(theEvent);
  
  // set kernel arguments
  status = clSetKernelArg(kernel,
			  0,
			  sizeof(cl_mem),
			  (void*) &d_scene);
  reportOpenCLerror(__LINE__, "Failed to set OpenCL arg. #1");
  
  status = clSetKernelArg(kernel,
			  1,
			  sizeof(cl_mem),
			  (void*) &d_pixelBuffer);
  reportOpenCLerror(__LINE__, "Failed to set OpenCL arg. #2");
  
  // Enqueue a kernel run call
  unsigned groupSize = workGroupSize;
  size_t globalThreads[1];
  globalThreads[0] = imgWidth * imgHeight;
  if (globalThreads[0] % groupSize != 0) {
    globalThreads[0] = (globalThreads[0] / groupSize + 1) * groupSize;
  }
  size_t localThreads[1];
  localThreads[0] = groupSize;
  
  status = clEnqueueNDRangeKernel(commandQueue,
				  kernel,
				  1,
				  NULL,
				  globalThreads,
				  localThreads,
				  0,
				  NULL,
				  &kernelEvent);
  reportOpenCLerror(__LINE__, "Failed to enqueue OpenCL kernel");
  
  animatePositions(scene);  // compute next anim state in parallel

  // Wait for the kernel call to finish execution
  status = clWaitForEvents(1, &kernelEvent);
  reportOpenCLerror(__LINE__, "Failed to wait for kernel completion");

  // determine kernel runtime
  cl_ulong time_start, time_end;
  
  clGetEventProfilingInfo(kernelEvent, CL_PROFILING_COMMAND_START,
			  sizeof(time_start), &time_start, NULL);
  clGetEventProfilingInfo(kernelEvent, CL_PROFILING_COMMAND_END,
			  sizeof(time_end), &time_end, NULL);

  timeIdx = (timeIdx + 1) & 31;
  total_time -= times[timeIdx];
  times[timeIdx] = time_end - time_start;
  total_time += times[timeIdx];

  throughput = 1e9/double(total_time >> 5);

  clReleaseEvent(kernelEvent);
  
  // copy result pixels from device to host
  status = clEnqueueReadBuffer(commandQueue,
			       d_pixelBuffer,
			       CL_FALSE,
			       0,
			       sizeof(cl_uchar4) * imgWidth * imgHeight,
			       bitmap->pixels,
			       0,
			       NULL,
			       &theEvent);
  reportOpenCLerror(__LINE__, "Failed to read from OpenCL output buffer");
  // wait for the copy operation to finish
  status = clWaitForEvents(1, &theEvent);
  reportOpenCLerror(__LINE__, "Failed to wait for read from OpenCL output");
  clReleaseEvent(theEvent);
}

void cleanUp() {
  delete bitmap;
  delete scene;
}


////////////////////////////////////////////
// static methods used for glut callbacks //
////////////////////////////////////////////

// quits when the escape key is pressed
void keyboardFunc(unsigned char key, int x, int y) {
  switch (key) {
  case 27:              // keycode for the Escape key
    cleanUp();
    printf("\n");
    exit(0);
    break;
  }
}

// helper routine for OpenGL text output
static void PrintString(void *font, const char *string) {
  int len, i;
  
  len = (int)strlen(string);
  for (i = 0; i < len; i++)
    glutBitmapCharacter(font, string[i]);
}

// calls computation and OpenGL draw when GLUT wants to refresh the display
void displayFunc(void) {
  static char buf[1024];
  
  computeImage(scene, bitmap->pixels);
  bitmap->draw();

  // superimpose some text: title, credits, acknowledgments, performance
  glColor3f(1.f, 1.f, 1.f);
  glRasterPos2f(-0.99f, 0.96f);
  PrintString(GLUT_BITMAP_HELVETICA_10, "jugCLer 1.0 by Holger Bettag "
	      "(hobold@vectorizer.org)");
  glRasterPos2f(0.09f, 0.96f);
  PrintString(GLUT_BITMAP_HELVETICA_10, "heavily inspired by Michael Birken's"
	      " reverse engineering of Eric Graham's original Amiga"
	      " juggler demo");  
  snprintf(buf, 1023, "theoretical peak: %6.2f fps", throughput);
  glRasterPos2f(-0.99f, -0.98f);
  PrintString(GLUT_BITMAP_HELVETICA_10, buf);  
  glRasterPos2f(0.4f, -0.98f);
  PrintString(GLUT_BITMAP_HELVETICA_10, "OpenCL boilerplate code based on"
	      " David Bucciarelli's MandelGPU");  

  glutSwapBuffers();
}

// keep the animation running as fast as possible with a GLUT idle callback
void idleFunc(void) {
  glutPostRedisplay();
}

/******************
 *  OpenCL Setup  *
 ******************/
// read kernel source code from file
char *ReadSources(const char *fileName) {  // taken from MandelGPU
  FILE *file = fopen(fileName, "r");
  if (!file) {
    fprintf(stderr, "Failed to open file '%s'\n", fileName);
    exit(-1);
  }
  
  if (fseek(file, 0, SEEK_END)) {
    fprintf(stderr, "Failed to seek file '%s'\n", fileName);
    exit(-1);
  }
  
  long size = ftell(file);
  if (size == 0) {
    fprintf(stderr, "Failed to check position on file '%s'\n", fileName);
    exit(-1);
  }
  
  rewind(file);
  
  char *src = (char *)malloc(sizeof(char) * size + 1);
  if (!src) {
    fprintf(stderr, "Failed to allocate memory for file '%s'\n", fileName);
    exit(-1);
  }
  
  fprintf(stderr, "Reading file '%s' (size %ld bytes)\n", fileName, size);
  size_t res = fread(src, 1, sizeof(char) * size, file);
  if (res != sizeof(char) * size) {
    fprintf(stderr, "Failed to read file '%s' (read %ld)\n", fileName, res);
    exit(-1);
  }
  src[size] = '\0'; // NULL terminated
  
  fclose(file);
  
  return src;
}

// find and prepare an OpenCL device
void setupOpenCL(void) {  // heavily inspired by MandelGPU
  cl_uint numPlatforms;
  cl_platform_id platform = NULL;
  status = clGetPlatformIDs(0, NULL, &numPlatforms);
  reportOpenCLerror(__LINE__, "Failed to get OpenCL platforms");
  
  if (numPlatforms > 0) {
    cl_platform_id *platforms =
      (cl_platform_id *)malloc(sizeof(cl_platform_id) * numPlatforms);
    status = clGetPlatformIDs(numPlatforms, platforms, NULL);
    reportOpenCLerror(__LINE__, "Failed to get OpenCL platform IDs");
    
    unsigned int i;
    for (i = 0; i < numPlatforms; ++i) {
      char pbuf[100];
      status = clGetPlatformInfo(platforms[i],
				 CL_PLATFORM_VENDOR,
				 sizeof(pbuf),
				 pbuf,
				 NULL);
      reportOpenCLerror(__LINE__, "Failed to get OpenCL platform info");
      
      status = clGetPlatformIDs(numPlatforms, platforms, NULL);
      reportOpenCLerror(__LINE__, "Failed to get OpenCL platform IDs");
      
      fprintf(stderr, "OpenCL Platform %d: %s\n", i, pbuf);
    }
    
    platform = platforms[0];
    free(platforms);
  }
  
  cl_context_properties cps[3] = {
    CL_CONTEXT_PLATFORM,
    (cl_context_properties) platform,
    0
  };
  
  cl_context_properties *cprops = (NULL == platform) ? NULL : cps;
  
  // first try to get a GPU device
  context = clCreateContextFromType(cprops,
				    CL_DEVICE_TYPE_GPU,
				    NULL,
				    NULL,
				    &status);
  if(status != CL_SUCCESS) {
    // try to fall back to CPU if GPU didn't work
    context = clCreateContextFromType(cprops,
				      CL_DEVICE_TYPE_CPU,
				      NULL,
				      NULL,
				      &status);
    reportOpenCLerror(__LINE__, "Failed to open OpenCL context");
  }

  // Get the size of device list data
  size_t deviceListSize;
  status = clGetContextInfo(context,
			    CL_CONTEXT_DEVICES,
			    0,
			    NULL,
			    &deviceListSize);
  reportOpenCLerror(__LINE__, "Failed to get OpenCL context info size");
  
  devices = (cl_device_id *) malloc(deviceListSize);
  if (devices == NULL) {
    fprintf(stderr, "Failed to allocate memory for OpenCL device list\n");
    exit(-1);
  }
  
  // Get the device list data
  status = clGetContextInfo(context,
			    CL_CONTEXT_DEVICES,
			    deviceListSize,
			    devices,
			    NULL);
  reportOpenCLerror(__LINE__, "Failed to get OpenCL context info");
  
  // Print devices list
  unsigned int i;
  for (i = 0; i < deviceListSize / sizeof(cl_device_id); ++i) {
    cl_device_type type = 0;
    status = clGetDeviceInfo(devices[i],
			     CL_DEVICE_TYPE,
			     sizeof(cl_device_type),
			     &type,
			     NULL);
    reportOpenCLerror(__LINE__, "Failed to get OpenCL device info");
    
    const char *stype;
    switch (type) {
    case CL_DEVICE_TYPE_ALL:
      stype = "TYPE_ALL";
      break;
    case CL_DEVICE_TYPE_DEFAULT:
      stype = "TYPE_DEFAULT";
      break;
    case CL_DEVICE_TYPE_CPU:
      stype = "TYPE_CPU";
      break;
    case CL_DEVICE_TYPE_GPU:
      stype = "TYPE_GPU";
      break;
    default:
      stype = "TYPE_UNKNOWN";
      break;
    }
    fprintf(stderr, "OpenCL Device %d: Type = %s\n", i, stype);
    
    char buf[512];
    status = clGetDeviceInfo(devices[i],
			     CL_DEVICE_NAME,
			     sizeof(char[512]),
			     &buf,
			     NULL);
    reportOpenCLerror(__LINE__, "Failed to get OpenCL device info");
    
    fprintf(stderr, "OpenCL Device %d: Name = %s\n", i, buf);
    
    status = clGetDeviceInfo(devices[i],
			     CL_DEVICE_VERSION,
			     sizeof(char[512]),
			     &buf,
			     NULL);
    reportOpenCLerror(__LINE__, "Failed to get OpenCL device info");
    
    fprintf(stderr, "OpenCL Device %d: OpenCL Version = %s\n", i, buf);
    
    status = clGetDeviceInfo(devices[i],
			     CL_DEVICE_EXTENSIONS,
			     sizeof(char[512]),
			     &buf,
			     NULL);
    reportOpenCLerror(__LINE__, "Failed to get OpenCL device info");
    
    fprintf(stderr, "OpenCL Device %d: Supported Extensions = %s\n", i, buf);
    
    cl_uint units = 0;
    status = clGetDeviceInfo(devices[i],
			     CL_DEVICE_MAX_COMPUTE_UNITS,
			     sizeof(cl_uint),
			     &units,
			     NULL);
    reportOpenCLerror(__LINE__, "Failed to get OpenCL device info");

    fprintf(stderr, "OpenCL Device %d: Compute units = %u\n", i, units);
    
    size_t gsize = 0;
    status = clGetDeviceInfo(devices[i],
			     CL_DEVICE_MAX_WORK_GROUP_SIZE,
			     sizeof (size_t),
			     &gsize,
			     NULL);
    reportOpenCLerror(__LINE__, "Failed to get OpenCL device info");
    
    fprintf(stderr, "OpenCL Device %d: Max. work group size = %u\n",
	    i, (unsigned int)gsize);
  }
  
  // Create the kernel program
  const char *sources = ReadSources(kernelFileName);
  program = clCreateProgramWithSource(context,
				      1,
				      &sources,
				      NULL,
				      &status);
  reportOpenCLerror(__LINE__, "Failed to open OpenCL kernel sources");

  status = clBuildProgram(program, 1, devices, NULL, NULL, NULL);
  if (status != CL_SUCCESS) {
    fprintf(stderr, "line %d: Failed to build OpenCL kernel\n", __LINE__);
    fprintf(stderr, "(CL error number %d: %s)\n", status,
	    openCLerrorNumberToString(status));
    
    
    size_t retValSize;
    status = clGetProgramBuildInfo(program,
				   devices[0],
				   CL_PROGRAM_BUILD_LOG,
				   0,
				   NULL,
				   &retValSize);
    reportOpenCLerror(__LINE__, "Failed to get OpenCL kernel info size");
    
    // Szaq's patch for NVIDIA OpenCL and Windows
    char *buildLog = (char *)malloc(retValSize + 1);
    status = clGetProgramBuildInfo(program,
				   devices[0],
				   CL_PROGRAM_BUILD_LOG,
				   retValSize,
				   buildLog,
				   NULL);
    reportOpenCLerror(__LINE__, "Failed to get OpenCL kernel info");
    buildLog[retValSize] = '\0';
    
    fprintf(stderr, "OpenCL Programm Build Log: %s\n", buildLog);
    exit(-1);
  }
  
  kernel = clCreateKernel(program, "render_gpu", &status);
  reportOpenCLerror(__LINE__, "Failed to create OpenCL kernel");
  
  // LordCRC's patch for better workGroupSize
  size_t gsize = 0;
  status = clGetKernelWorkGroupInfo(kernel,
				    devices[0],
				    CL_KERNEL_WORK_GROUP_SIZE,
				    sizeof (size_t),
				    &gsize,
				    NULL);
  reportOpenCLerror(__LINE__, "Failed to get OpenCL kernel work group info");
  
  workGroupSize = (unsigned int) gsize;
  fprintf(stderr, "OpenCL Device 0: kernel work group size = %d\n",
	  workGroupSize);

  // we want timing information to be recorded in this command queue
  cl_command_queue_properties prop = CL_QUEUE_PROFILING_ENABLE;
  commandQueue = clCreateCommandQueue(context,
				      devices[0],
				      prop,
				      &status);
  reportOpenCLerror(__LINE__, "Failed to create OpenCL command queue");  
  
  const int numPixels = imgWidth * imgHeight;
  cl_uint sizeInBytes = sizeof(cl_uchar4) * numPixels;
  d_pixelBuffer = clCreateBuffer(context,
				 CL_MEM_WRITE_ONLY,
				 sizeInBytes,
				 NULL,
				 &status);
  reportOpenCLerror(__LINE__, "Failed to create OpenCL output buffer");  
  
  d_scene = clCreateBuffer(context,
			   CL_MEM_READ_ONLY,
			   sizeof(Scene),
			   NULL,
			   &status);
  reportOpenCLerror(__LINE__, "Failed to create OpenCL input buffer");  
}

/********************************************
 *  setup GLUT and OpenCL, start animation  *
 ********************************************/
int main(int argc, char **argv) {
  // initialize GLUT, open display window, and register callback functions
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA);
  glutInitWindowSize(imgWidth, imgHeight);
  glutCreateWindow("jugCLer 1.0 - \"A Classic Reanimated\" - "
		   "press ESC to close");
  glutKeyboardFunc(keyboardFunc);
  glutDisplayFunc(displayFunc);
  glutIdleFunc(idleFunc);

  // set up buffers
  bitmap = new Bitmap(imgWidth, imgHeight);
  scene = new Scene;
  if ((bitmap == NULL) || (scene == NULL)) {
    fprintf(stderr, "out of memory for scene, aborting\n");
    exit(EXIT_FAILURE);
  }
  
  // set up OpenCL
  setupOpenCL();
  
  // set up scene
  setupAnim(scene, imgWidth, imgHeight);

  // Action!
  glutMainLoop();
  
  exit(EXIT_SUCCESS);
}
